import socket
import socketserver
import threading
import time
import logging
import json
import struct
logger = logging.getLogger(__name__)



class MonitorServer:

  def __init__(self, local_hostname, port):
    self.connections = []
    self.local_hostname = local_hostname
    self.port = port
    self.server = None
    self.listener_thread = None

  def listen_for_new_clients(self):
    ms = self
    class NewClientHandler(socketserver.BaseRequestHandler):
        def handle(self):
            # self.request is the TCP socket connected to the client
            self.data = self.request.recv(1024).strip()
            s = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
            s.connect( (self.client_address[0], int(self.data)) )
            ms.connections.append(s)
            logger.info(f"Client connected: '{self.client_address}' on port '{self.data}'")

    logger.info(f"Setting up TCP server to listed for client connections on port {self.port} as {self.local_hostname}.")
    with socketserver.TCPServer((self.local_hostname, self.port), NewClientHandler) as server:
        self.server = server
        server.serve_forever()

  def broadcast_message(self,message):
    if not isinstance(message,(str,bytes)):
      message = json.dumps(message)

    if logger.isEnabledFor(logging.DEBUG):
      logger.debug(f"Broadcasting message: '{message}'")

    if not isinstance(message,bytes):
      message = message.encode('utf-8')

    message = struct.pack('>I',len(message)) + message

    live_connections = []
    for client in self.connections:
      try:
        client.sendall(message)
        if logger.isEnabledFor(logging.DEBUG):
          logger.debug(f"Actual message sent to {client}: '{message}'")
        live_connections.append(client)
      except BrokenPipeError:
        client.close()
      except ConnectionResetError:
        client.close()
    self.connections = live_connections


  def start(self):
    self.listener_thread = threading.Thread(target=self.listen_for_new_clients)
    self.listener_thread.start()

  def shutdown(self):
    if self.server:
      self.server.shutdown()
      self.server = None
    if self.listener_thread:
      self.listener_thread.join()
      self.listener_thread = None



import os
import select
import threading
import logging


try:
  os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
  import pygame
  have_pygame = True
except:
  have_pygame = False

logger = logging.getLogger(__name__)

class PygameMessageDisplay:
  def __init__(self,width=800,height=20,font_name="courier",font_size=20):
    self.text_font_name = font_name
    self.text_font_size = font_size
    self.text_color = (200,200,200)
    self.win_width = width
    self.win_height = height
    self.win_color = (0,0,0)



    self.running = False
    self.exit = False
    self.piper,self.pipew = os.pipe()
    self.screen = None

    self.display_thread = threading.Thread(target = self.run)



  def shutdown(self):
    self.exit = True
    self.display_thread.join()
    pygame.quit()

  def start(self):
    logger.info("Initializing pygame")
    pygame.font.init()
    pygame.display.init()
    logger.info("Starting display thread")
    self.display_thread.start()



  def run(self):
    self.running = True
    self.text_font = pygame.font.SysFont(self.text_font_name,self.text_font_size)
    self.screen = pygame.display.set_mode((self.win_width,self.win_height),pygame.NOFRAME)
    self.screen.fill(self.win_color)
    pygame.display.flip()
    self.set_message("")

    while not self.exit:
      ifds,ofds,efds = select.select([self.piper],[],[],1)
      if len(ifds) > 0:
        self.screen.fill(self.win_color)

        text = os.read(ifds[0],2048).decode()
        rendered_text = self.text_font.render(text,True,self.text_color)
        pos = ( (self.screen.get_width()- rendered_text.get_width())/2, rendered_text.get_height())

        self.screen.blit(rendered_text, pos )
        pygame.display.flip()

  def set_message(self,text):
    os.write(self.pipew,text.encode())

  def is_running(self):
    return self.running

class DummyMessageDisplay:
  def __init__(self,width=800,height=200,font_name="courier",font_size=20):
    pass

  def start(self):
    pass

  def shutdown(self):
    pass

  def set_message(self,text):
    pass

  def is_running(self):
    return True

if have_pygame:
  MessageDisplay = PygameMessageDisplay
else:
  MessageDisplay = None



import socket
import select
import struct
import time
import logging
logger = logging.getLogger(__name__)

class MonitorClient:
  def __init__(self,remote_hostname,local_hostname,port_range):
    self.remote_hostname = remote_hostname
    self.local_hostname = local_hostname
    self.port_range = port_range
    self.exit = False
    self.slots = []

  def shutdown(self):
    self.exit = True

  def add_slot(self,f):
    self.slots.append(f)

  @staticmethod
  def recvn(fd,n):
    data = bytearray(n)
    i = 0
    while i < n:
      packet = fd.recv(n - i)
      if not packet:
        return None
      data[i:i+len(packet)] = packet
      i += len(packet)
    return data

  @staticmethod
  def recv_msg(fd):
    # read message length
    data = MonitorClient.recvn(fd,4)
    if not data or len(data) < 1:
      return None
    N = struct.unpack('>I', data)[0]
    if logger.isEnabledFor(logging.DEBUG):
      logger.debug(f"Recieved message prefix: '{data}' indicates we need to read {N} bytes")
    # read message
    data = MonitorClient.recvn(fd,N)
    if logger.isEnabledFor(logging.DEBUG):
      logger.debug(f"Recieved message body: '{data}'")

    if not data or len(data) < 1:
      return None

    return bytes(data)

  def start(self):
    # we need to:
    # 1. find a port we can listen on.
    # 2. bind a socket to the port.
    # 3. connect to server and send it the port we are listening on.
    # 4. wait for, and process, messages from server.
    for port in range(*self.port_range):
      # find open port
      with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
          s.bind( (self.local_hostname,port) )
        except:
          continue
        # we have a port with a socket attached.
        while not self.exit:

          # start listening for connections
          s.listen()
          # connect to server and send them our port.
          # we might have to wait.
          with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as ss:
            while ss.connect_ex( (self.remote_hostname,3000) ) != 0:
              time.sleep(1)
              if self.exit:
                return
            ss.sendall( str(port).encode('utf-8') )

          # process messages from server
          connection,gsc_address = s.accept()
          while not self.exit:
            ifds,ofds,efds = select.select([connection],[],[],1)
            if len(ifds) > 0:
              # read message length
              message = MonitorClient.recv_msg(connection)
              if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f"Recieved message: '{message}'")
              if message is None:
                break
              for slot in self.slots:
                slot(message)

          # s.shutdown(socket.SHUT_RDWR)
          if self.exit:
            return



from pathlib import Path
import logging
import threading
import time
import json
import os,sys
import select


from .ScriptRecorder import *
from .ScriptedSession import *
from .MonitorServer import *
from .MonitorClient import *
from . import ucode

import click

def make_interactive_shell(shell):
    oshell = shell
    if shell is None:
        shell = os.environ.get("SHELL", "bash")

    shell = shell.split()
    shell[0] = shutil.which(shell[0])

    if shell[0] is None:
      raise click.RuntimeError("Could not find working shell for '{oshell}'. Check that it is in your path.")

    return shell


@click.command(help="Run a shell script interactivly.")
@click.argument("script",type=click.Path(exists=True))
@click.option("--shell","-s",help="The shell to use for running the script.")
@click.option("--debug","-d",is_flag=True,help="Log debug messages.")
@click.option("--verbose","-v",is_flag=True,help="Log info messages.")
@click.option("--no-statusline/--statusline",help="disable/enable status line (the statusline may interfere with some full screen applications, such as vim). status line is enabled by default.")
@click.option("--line-mode","-l",is_flag=True,help="Start script in line mode.")
@click.option("--startup-command", multiple=True, help="Send TEXT to the terminal before starting input (a \\r will be appended). May be given multiple times.")
@click.option("--monitor/--no-monitor", multiple=True, help="Run a server to send status information to gsc monitor clients (such as gsc-monitor).")
def gsc(script,shell,debug,verbose,no_statusline,line_mode,startup_command,monitor):

    logger = None
    if verbose or debug:
      logger = logging.getLogger()
      fh = logging.FileHandler("gsc.log")
      fmt = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
      fh.setFormatter(fmt)
      logger.addHandler(fh)
      if verbose:
        logger.setLevel(logging.INFO)
      if debug:
        logger.setLevel(logging.DEBUG)

    shell = make_interactive_shell(shell)
    if logger:
      logger.debug(f"Shell: {shell}")
    session = ScriptedSession(script,shell)
    session.set_statusline(not no_statusline)
    session.set_monitor(monitor)
    if line_mode:
      session.mode = session.Modes.Line
    try:
      for command in startup_command:
        session.send_to_terminal(command+"\r")
      session.run()
    finally:
      session.cleanup()

    if logger:
      logger.debug("Session finished")

    return 0



@click.command(help="Record gsc script.")
@click.argument("output",type=click.Path(exists=False))
@click.option("--shell","-s",help="The shell to use for running the script.")
def gsc_record(output,shell):
  output = pathlib.Path(output)

  shell = make_interactive_shell(shell)
  rec = ScriptRecorder(shell)
  rec.run()
  rec.cleanup()

  output.write_bytes(rec.get_bytes())
  print()
  print("Script recording finished.")



@click.command(help="Display keycodes for keypresses.")
@click.option("--keypress-driver","-k",is_flag=True,help="Use keypress event driver instead of characters from stdin.")
def display_keycodes(keypress_driver):
  import tty, termios, sys, os
  import collections

  try:
    from pynput.keyboard import Key,KeyCode,Listener
    have_pynput = True
  except:
    have_pynput = False

  
  if keypress_driver:
    if not have_pynput:
      print("Could not run keypress-based driver. `pynput` is not installed.")
      sys.exit(1)

    def on_press(key):
      print("stroke: down")
      print("raw:",key)
      print("type:",type(key))
      if type(key) == KeyCode:
        print("dead:",key.is_dead)
        print("char:",key.char)
      else:
        print("val:",key.value)
      print()

    def on_release(key):
      print("stroke: up")
      print("raw:",key)
      print("type:",type(key))
      if type(key) == KeyCode:
        print("dead:",key.is_dead)
        print("char:",key.char)
      else:
        print("val:",key.value)
      print()

      if key == Key.esc:
        return False

    with Listener(on_press=on_press,on_release=on_release) as listener:
      listener.join()

    return



  saved = termios.tcgetattr(sys.stdin.fileno())
  tty.setraw(sys.stdin.fileno())
  while True:
    input = os.read(sys.stdin.fileno(),1024)
    print(len(input),'chars',end='\n\r')
    print("raw:",f"`{input}`",end="\n\r")
    print("unicod:",f"`{input.decode(ucode)}`",end="\n\r")
    sys.stdout.write("int: ")
    for c in input:
      sys.stdout.write(f"{c} ")
    print(end="\n\r")
    sys.stdout.write("hex: ")
    for c in input:
      sys.stdout.write(f"{hex(c)} ")
    print(end="\n\r")
    print(end="\n\r")

    if len(input) == 1 and ord(input) == 4:
      break


  termios.tcsetattr(sys.stdin.fileno(),termios.TCSANOW,saved)


@click.command(help="A gsc monitor client to status information about the current gsc session.")
@click.option("--remote-hostname","-r",default='localhost',help="The remote server (running gsc) hostname.")
@click.option("--local-hostname","-l",default='localhost',help="The local server (running gsc-monitor) hostname.")
@click.option("--port","-p",help="Specify the (local) port to use for communicating with gsc session.")
@click.option("--debug","-d",is_flag=True,help="Log debug messages.")
@click.option("--verbose","-v",is_flag=True,help="Log info messages.")
def gsc_monitor(remote_hostname,local_hostname,port,debug,verbose):
  logger = None
  if verbose or debug:
    logger = logging.getLogger()
    fh = logging.FileHandler("gsc-monitor.log")
    fmt = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    fh.setFormatter(fmt)
    logger.addHandler(fh)
    if verbose:
      logger.setLevel(logging.INFO)
    if debug:
      logger.setLevel(logging.DEBUG)

  port_range = (3001,3020)
  if port:
    port_range = (port,port)



  try:
    import urwid
  except:
    print("ERROR: could not import 'urwid' module. Please install it and try again.")
    sys.exit(1)


  piper,pipew = os.pipe()
  gsc_monitor = MonitorClient(remote_hostname,local_hostname,port_range)
  gsc_monitor.add_slot(lambda data: os.write(pipew,data))



  def handle_input(char):
    if char in ('q','Q'):
      raise urwid.ExitMainLoop()

  def message_handler():
    message = os.read(piper,8096)
    try:
      state = json.loads(message.decode('utf-8'))
    except:
      logger.error("Could not decode message: {message}")
      return
    l,c = state['pos']
    completed_lines = '\n'.join(state['lines'][0:l])+'\n'
    current_line_completed_chars = state['lines'][l][:c]
    current_line_upcomming_chars = state['lines'][l][c:]+'\n'
    upcomming_lines = '\n'.join(state['lines'][l+1:])+'\n'
    text = []
    text.append( ('dg',completed_lines) )
    text.append( ('dr',current_line_completed_chars) )
    text.append( ('lr',current_line_upcomming_chars) )
    text.append( ('lg',upcomming_lines) )
    body.set_text(text)


  # print(urwid.get_all_fonts())
  body = urwid.Text("")
  main_frame = urwid.Frame(body=urwid.Filler(body), header=urwid.Text("gsc Monitor"))


  palette = [
      ('divider','','','','g27','#a06'),
      ('lg','light gray','default'),
      ('dg','dark gray','default'),
      ('dr','dark red','default'),
      ('lr','light red','default'),
      ]
  loop = urwid.MainLoop(main_frame,palette=palette,unhandled_input=handle_input)
  loop.watch_file(piper,message_handler)
  # loop.set_alarm_in(0.1,poll,None)
  
  client_thread = threading.Thread(target=gsc_monitor.start)
  client_thread.start()
  loop.run()
  gsc_monitor.shutdown()
  client_thread.join()





@click.command(help="A gsc monitor test client to test gsc monitor servers.")
@click.option("--remote-hostname","-r",default='localhost',help="The remote server (running gsc) hostname.")
@click.option("--local-hostname","-l",default='localhost',help="The local server (running gsc-monitor) hostname.")
@click.option("--port","-p",help="Specify the (local) port to use for communicating with gsc session.")
def gsc_monitor_test_client(remote_hostname,local_hostname,port):
  logger = logging.getLogger()
  logger.setLevel(logging.INFO)
  ch = logging.StreamHandler()
  logger.addHandler(ch)
  port_range = (3001,3020)
  if port:
    port_range = (port,port)


  gsc_monitor = MonitorClient(remote_hostname,local_hostname,port_range)
  gsc_monitor.add_slot(lambda data: print(f"Recieved: '{data}'"))
  thread = threading.Thread(target=gsc_monitor.start)
  thread.start()
  print("Press enter to shutdown:")
  exit = False
  while not exit:
    ready = select.select([sys.stdin],[],[],1)[0]
    if ready:
      exit = True
  gsc_monitor.shutdown()
  thread.join()




@click.command(help="Run a gsc monitor server to test gsc monitor clients.")
@click.option("--local-hostname","-l",default="localhost",help="The server hostname.")
@click.option("--port","-p",default=3000,help="Specify the (local) port to lisiten for new clients on.")
def gsc_monitor_test_server(local_hostname,port):
  logger = logging.getLogger()
  logger.setLevel(logging.INFO)
  ch = logging.StreamHandler()
  logger.addHandler(ch)
  server = MonitorServer(local_hostname,port)
  server.start()
  print("Press enter to shutdown:")
  exit = False
  while not exit:
    ready = select.select([sys.stdin],[],[],1)[0]
    if ready:
      exit = True
    else:
      mesg = {'desc' : 'test message', 'mode' : 'Modes.Insert', 'lines' : ["pwd","ls","cat tmp.txt","gnuplot","exit"], 'pos':(2,5)}
      server.broadcast_message(mesg)

  server.shutdown()



from .Script import *
from .TerminalSession import *
from .UserInputHandler import *
from .CommandParser import *
from .MonitorServer import *
from .MessageDisplay import *
from . import ucode
import sys,tty,string
import enum
import logging
try:
  import blessings
  have_blessings = True
except:
  have_blessings = False

logger = logging.getLogger(__name__)

class ScriptedSession:


    class Mode:
      def __init__(self,session):
        self.session = session

    class InsertMode(Mode):
      def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)

      def run(self):
        logger.debug("entering insert mode")
        while not self.session.script.eof():
          if self.handle_script_line(): return
          while not self.session.script.eol():
            if self.handle_script_current_char(): return
            self.session.update_statusline()
          if self.handle_script_eol(): return
          self.session.update_statusline()
        if self.handle_script_eof(): return
        self.session.update_statusline()

        self.session.exit_flag = True

      def handle_script_line(self):
        '''
        Look at current script line and
        take any actions necessary.

        Returns True if session to continue to next line.
        '''
        cl = self.session.script.current_line()
        if cl is None:
          return False
        res = self.session.command_parser.parse(cl)
        if res is None:
          return False

        if res['name'] == "comment":
          return False



        if res['name'] == "passthrough":
          self.session.mode = self.session.Modes.Passthrough
          self.session.script.seek_next_line()
          logger.debug(f"switching modes: insert -> pass-through")
          return True

        if res['name'] == "temporary passthrough":
            logger.debug(f"temporarily switching to pass-through mode")
            self.session.mode = self.session.Modes.TemporaryPassthrough
            self.session.script.seek_next_line()
            self.session.mode_handlers[self.session.mode].run()
            self.session.mode = self.session.Modes.Insert

        if res['name'] == "statusline":
          if len(res['args']) < 1:
            self.session.set_statusline(True)
          else:
            if res['args'][0].lower() in ["off","false","0","disable"]:
              self.session.set_statusline(False)
            elif res['args'][0].lower() in ["on","true","1","enable"]:
              self.session.set_statusline(True)
            else:
              logger.error(f"Unrecognized value '{res['args'][0]}' in statusline command. Ignoring.")
          self.session.script.seek_next_line()

        if res['name'] == "display":
          if self.session.message_display is None:
            if MessageDisplay is None:
              logger.error(f"Script contained a 'display' command, but no message displayer could be found. Make sure you have one of the supported GUI kits installed (pygame for example).")
              self.session.message_display = DummyMessageDisplay()
            else:
              self.session.message_display = MessageDisplay()

          if not self.session.message_display.is_running():
            self.session.message_display.start()

          if len(res['args']) > 0:
            self.session.message_display.set_message(res['args'][0])
          self.session.script.seek_next_line()

          return True
        
        if res['name'] == "pause":
          self.session.script.seek_next_line()
          if len(res['args']) > 0:
            t = float(res['args'][0])
            if t < 0:
              logger.info(f"Pausing until user presses a key.")
              self.session.input_handler.read()
            else:
              logger.info(f"Pausing for {t} seconds...")
              time.sleep(t)
          return True


        return False


      def handle_script_current_char(self):
        '''
        Read input from user and decide what to
        do with the current script character.
        '''
        input = self.session.input_handler.read()
        if self.session.input_handler.last_read_ord in [127]: # backspace
          self.session.terminal.send(chr(127))
          self.session.script.seek_prev_col()
        elif self.session.input_handler.last_read_ord in [4]: # ctl-d
          logger.debug(f"switching modes: insert -> command")
          self.session.mode = self.session.Modes.Command
          return True
        elif self.session.input_handler.last_read_ord in [3]: # ctl-c
          self.session.exit_flag = True
          return True
        else:
          self.session.terminal.send(self.session.script.current_char())
          self.session.script.seek_next_col()

        return False

        
      def handle_script_eol(self):
        '''
        Read input from user and decide what to
        do at the end of the line.
        '''
        while True:
          input = self.session.input_handler.read().decode(ucode)
          if self.session.input_handler.last_read_ord in [127]: # backspace
            self.session.terminal.send(chr(127))
            self.session.script.seek_prev_col()
            return False

          if self.session.input_handler.last_read_ord in [4]: # ctl-d
            logger.debug(f"switching modes: insert -> command")
            self.session.mode = self.session.Modes.Command
            return True

          if self.session.input_handler.last_read_ord in [3]: # ctl-c
            self.session.exit_flag = True
            return True

          if self.session.input_handler.last_read_ord in [13]: # return
            self.session.terminal.send('\r')
            self.session.script.seek_next_line()
            return False



      def handle_script_eof(self):
        '''
        Read input from user and decide what to
        do at the end of the file.
        '''
        while True:
          input = self.session.input_handler.read().decode(ucode)
          if self.session.input_handler.last_read_ord in [13]: # return
            logger.debug(f"recieved user input {input}. exitting")
            return False
          elif self.session.input_handler.last_read_ord in [4]: # ctl-4
            logger.debug(f"switching modes: insert -> command")
            self.session.mode = self.session.Modes.Command
            return True
          else:
            logger.debug(f"recieved user input {input} (ord {self.session.input_handler.last_read_ord}), but did not exit")


    class LineMode(InsertMode):
      def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)

      def handle_script_current_char(self):
        '''
        Overload InserMode single char handler to send entire lines.
        '''
        line = self.session.script.current_line()
        self.session.script.seek_end_col()
        self.session.terminal.send(line)
        return False





    class CommandMode(Mode):
      def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)

      def run(self):
        logger.debug("entering command mode")
        self.handle_user_input()
        logger.debug("exiting command mode")

      def handle_user_input(self):
        while True:
          input = self.session.input_handler.read().decode(ucode)
          if input in ["i"]:
            logger.debug(f"switching modes: command -> insert")
            self.session.mode = self.session.Modes.Insert
            return
          if input in ["I"]:
            logger.debug(f"switching modes: command -> line")
            self.session.mode = self.session.Modes.Line
            return
          if input in ["p"]:
            logger.debug(f"switching to pass-through mode")
            self.session.mode = self.session.Modes.Passthrough
            return
          if input in ["P"]:
            logger.debug(f"temporarily switching to pass-through mode")
            self.session.mode = self.session.Modes.TemporaryPassthrough
            self.session.mode_handlers[self.session.mode].run()
            self.session.mode = self.session.Modes.Command
          if input in ["q"]:
            self.session.exit_flag = True
            return
          if input in ["s"]:
            self.session.toggle_statusline()
            return
          if input in ["h"]:
            self.session.script.seek_prev_col()
          if input in ["j"]:
            self.session.script.seek_next_line(ret=False)
          if input in ["k"]:
            self.session.script.seek_prev_line(ret=False)
          if input in ["l"]:
            self.session.script.seek_next_col()
          if input in ["^"]:
            self.session.script.seek_beg_col()
          if input in ["$"]:
            self.session.script.seek_end_col()

          self.session.update_statusline()


    class PassthroughMode(Mode):
      def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)

      def run(self):
        logger.debug("entering pass-through mode")
        self.session.update_statusline()
        while True:
          input = self.session.input_handler.read().decode(ucode)
          if self.session.input_handler.last_read_ord in [4]: # ctl-d
            logger.debug("switching mode: pass-through -> command")
            self.session.mode = self.session.Modes.Command
            return
          self.session.terminal.send(input)
          self.session.update_statusline()


    class TemporaryPassthroughMode(Mode):
      def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)

      def run(self):
        logger.debug("entering temporary pass-through mode")
        self.session.update_statusline()
        while True:
          input = self.session.input_handler.read().decode(ucode)
          self.session.terminal.send(input)
          self.session.update_statusline()
          if self.session.input_handler.last_read_ord in [13]: # return
            logger.debug("exiting temporary pass-through mode")
            return






    Modes = Enum("Modes",'Insert Line Command Passthrough TemporaryPassthrough')
    def __init__(self, script, shell):
        self.script = Script(script)
        self.shell = shell
        self.STDINFD = sys.stdin.fileno()
        self.exit_flag = False
        self.statusline_flag = False
        self.terminal = TerminalSession(self.shell)
        self.terminal.add_output_callback( lambda c: self.update_statusline() )
        self.saved_terminal_settings = None
        self.command_parser = CommandParser()

        self.command_parser.add_command("passthrough", ['pass-through','pass'])
        self.command_parser.add_command("temporary passthrough",['temporary pass-through', 'temp pass'])
        self.command_parser.add_command("line")
        self.command_parser.add_command("mode")
        self.command_parser.add_command("statusline")
        self.command_parser.add_command("stdout")
        self.command_parser.add_command("pause")
        self.command_parser.add_command("comment")
        self.command_parser.add_command("display")

        if have_blessings:
          self.bterm = blessings.Terminal()
          # print(self.bterm.clear())
        else:
          self.bterm = None

        try:
          self.saved_terminal_settings = termios.tcgetattr(self.STDINFD)
          tty.setraw(self.STDINFD)
        except:
          logging.warning("Could not save terminal state for stdin. This likely means that no terminal is in control.")

        self.input_handler = UserInputHandler(self.STDINFD)

        self.mode = self.Modes.Insert
        self.mode_handlers = {self.Modes.Insert:self.InsertMode(self),
                              self.Modes.Line:self.LineMode(self),
                              self.Modes.Command:self.CommandMode(self),
                              self.Modes.Passthrough:self.PassthroughMode(self),
                              self.Modes.TemporaryPassthrough:self.TemporaryPassthroughMode(self),
                              }
        self.mode_status_abbrvs = {self.Modes.Insert:"I",
                                   self.Modes.Command:"C",
                                   self.Modes.Passthrough:"P",
                                   self.Modes.TemporaryPassthrough:"T",
                                   self.Modes.Line:"L"}


        self.monitor_server = None
        self.monitor_server_flag = False
        self.monitor_server_hostname = "localhost"
        self.monitor_server_port = 3000

        self.message_display = None

    def cleanup(self):
      try:
        logger.debug("trying to stop terminal session")
        self.terminal.stop()
      except Exception as e:
        logger.debug(f"there was en error: {e}")

      if self.saved_terminal_settings is not None:
        try:
          logger.debug("trying to restore terminal settings")
          termios.tcsetattr(self.STDINFD, termios.TCSANOW, self.saved_terminal_settings)
        except Exception as e:
          logger.debug(f"there was en error: {e}")

      if self.message_display is not None and self.message_display.is_running():
        self.message_display.shutdown()
      self.message_display = None


    def send_to_terminal(self,text):
      self.terminal.send(text)

    def disable_terminal_output(self):
      self.terminal.set_output_mode(self.terminal.OutputMode.Drop)

    def enable_terminal_output(self):
      self.terminal.set_output_mode(self.terminal.OutputMode.Print)

    def set_terminal_output_mode(self,mode):
      self.terminal.set_output_mode(mode)


    def run(self):
      if self.monitor_server_flag:
        self.start_monitor_server()
      while not self.exit_flag:
        self.update_statusline()
        self.mode_handlers[self.mode].run()

      self.stop_monitor_server()
      logger.debug("Exiting session")


    def set_statusline(self,flag):
      self.statusline_flag = flag

    def toggle_statusline(self):
      self.statusline_flag = not self.statusline_flag

    def update_statusline(self):
      if self.statusline_flag:
        mode = self.mode_status_abbrvs[self.mode]
        lno = self.script.line + 1
        lNo = len(self.script.lines)
        cno = self.script.col + 1
        cNo = len(self.script.current_line(""))

        sline = f" {mode} {lno:03}/{lNo:03} {cno:03}/{cNo:03}"

        if self.bterm is not None:
          t = self.bterm
          os.write( self.terminal.STDOUTFD, (t.save() + t.move(0,t.width-len(sline)) + sline + t.restore()).encode(ucode) )

      if self.monitor_server:
        self.send_state_to_monitors()

      
    def set_monitor(self,flag):
      self.monitor_server_flag = flag

    def toggle_monitor(self):
      self.monitor_server_flag = not self.monitor_server_flag

    def start_monitor_server(self):
      self.monitor_server = MonitorServer(self.monitor_server_hostname,self.monitor_server_port)
      self.monitor_server.start()

    def stop_monitor_server(self):
      if self.monitor_server:
        self.monitor_server.shutdown()
        self.monitor_server = None

    def send_state_to_monitors(self):
      state = {'mode':str(self.mode),
          'pos': (self.script.line,  self.script.col),
          'lines': self.script.lines
          }
      self.monitor_server.broadcast_message(state)


    def set_message_display(self,display):
      self.message_display = display







import sys,os
import termios,tty
from queue import Queue
import logging

logger = logging.getLogger(__name__)


class UserInputHandler:
  def __init__(self,inputfd):
    self.inputfd = inputfd
    self.chunk_size = 1024
    self.last_read_input = None
    self.filter_input = False
    self.filters = []
    self.callbacks = []
    self.queue = Queue()


  def clear_filters(self):
    self.filters = []

  def add_filter( self, f ):
    self.filters.append(f)

  def add_callback( self, c ):
    self.callbacks.append(c)


  def filter(self, input):
    if self.filter_input:
      for f in self.filters:
        input = f(input)

    if logger.isEnabledFor(logging.DEBUG):
      logger.debug(f"filtered input to '{input}' ({len(input)} chars)")

    return input

  def run_callbacks(self):
    for c in self.callbacks:
      c()

  def queue_input(self,input):
    self.queue.put(input)

  def read(self):
    if self.queue.empty():
      self.last_read_input = os.read(self.inputfd,self.chunk_size)
    else:
      self.last_read_input = self.queue.get()

    if logger.isEnabledFor(logging.DEBUG):
      logger.debug(f"read '{self.last_read_input}' ({len(self.last_read_input)} chars)")

    self.run_callbacks()

    return  self.filter(self.last_read_input)

  @property
  def last_read(self):
    return self.filter(self.last_read_input)

  @property
  def last_read_ord(self):
    '''Return ordinal for last read input. If no ordinal exists, returns -1'''
    try:
      return ord(self.last_read)
    except:
      return -1













from pathlib import Path


class Script:
  def __init__(self,filename=None):
    self.filenames = []
    self.lines = []
    self.add_lines(filename)
    self.line = 0
    self.col = 0

  def eol(self):
    return self.line >= len(self.lines) or self.col >= len(self.lines[self.line])

  def eof(self):
    return self.line >= len(self.lines) or (self.line >= len(self.lines)-1 and self.col >= len(self.lines[self.line]))

  def add_lines(self,lines):

    if isinstance(lines,str):
      lines = Path(lines)

    if isinstance(lines,Path):
      self.filenames = str(lines)
      lines = lines.read_text().split("\n")
    else:
      self.filenames = "None"

    self.lines += lines

  def seek_next_line(self,n=1,ret=True):
    self.line += n
    self.line = min(self.line,len(self.lines))
    if ret:
      self.reset_seek_col()

  def seek_prev_line(self,n=1,ret=True):
    self.line -= n
    self.line = max(self.line,0)
    if ret:
      self.reset_seek_col()

  def seek_next_col(self,n=1):
    self.col += n
    self.col = min(self.col,len(self.current_line("")))

  def seek_prev_col(self,n=1):
    self.col -= n
    self.col = max(self.col,0)

  def seek_end_col(self):
    self.col = len(self.current_line(""))

  def seek_beg_col(self):
    self.col = 0

  def iter_lines(self):
    while self.line < len(self.lines):
      self.line += 1
      yield self.lines[self.line-1]

  def iter_chars_on_current_line(self):
    while self.line < len(self.lines) and self.col < len(self.lines[self.line]):
      self.col += 1
      yield self.lines[self.line][self.col-1]



  def current_line(self,return_on_out_of_range=None):
    N = len(self.lines)
    if self.line >= N or self.line < 0:
      return return_on_out_of_range
    return self.lines[self.line]

  def current_char(self):
    return self.lines[self.line][self.col]
      
  def reset_seek_line(self):
    self.line = 0

  def reset_seek_col(self):
    self.col = 0

  def reset_seek(self):
    self.reset_seek_line()
    self.reset_seek_col()









import tty
import pty
import os
import sys
import select
import termios
import threading
import time
import pathlib
import shutil
import signal
import fcntl
import struct
from queue import Queue
from enum import Enum
from . import ucode



class TerminalSession:
  class OutputMode(Enum):
    Print = 1
    Pause = 2
    Drop = 3


  def __init__(self,shell:list):

    self.output_mode = TerminalSession.OutputMode.Print
    self.STDINFD = sys.stdin.fileno()
    self.STDOUTFD = sys.stdout.fileno()
    self.STDERRFD = sys.stderr.fileno()
    self.shutdown_flag = False

    self.slavePID,self.termfd = pty.fork()

    if self.slavePID == pty.CHILD:
      os.execvp(shell[0], shell)

    # child proc will not make it to this point
    self.sync_windows_size()

    self.terminal_output_thread = threading.Thread(target=self._process_terminal_output)
    self.terminal_output_thread.start()

    self.output_callbacks = []


  def add_output_callback(self,f):
    self.output_callbacks.append(f)


  def _process_terminal_output(self):
    poll = select.poll()
    poll.register(self.termfd,select.POLLIN)

    stash = Queue()

    while not self.shutdown_flag:
      res = poll.poll(100)
      for fd,evm in res:
        if evm & select.POLLIN:
          c = os.read(fd,1024)
          if self.output_mode == TerminalSession.OutputMode.Drop:
            continue

          if self.output_mode == TerminalSession.OutputMode.Print:
            while not stash.empty():
              os.write(self.STDOUTFD,stash.get())
            os.write(self.STDOUTFD,c)

          if self.output_mode == TerminalSession.OutputMode.Pause:
            stash.put(c)

          for f in self.output_callbacks:
            f(c)
            

  def stop(self):
    os.kill(self.slavePID, signal.SIGKILL)
    self.shutdown_flag = True
    self.terminal_output_thread.join()


  def send(self,c : str):
    if not isinstance(c,bytes):
      c = c.encode(ucode)
    os.write(self.termfd,c)

  def sync_windows_size(self):
    winsz = struct.pack('HHHH',0,0,0,0)
    winsz = fcntl.ioctl(self.STDINFD, termios.TIOCGWINSZ, winsz)
    fcntl.ioctl(self.termfd, termios.TIOCSWINSZ, winsz)


  
  def set_output_mode(self,mode : OutputMode):
    self.output_mode = mode





import logging
logger = logging.getLogger(__name__)


try:
  from pyparsing import *
  have_pyparsing = True
except:
  logger.warning("pyparsing module could not be imported. Command parsing ability will be (severely) limited")
  have_pyparsing = False



class CommandParser:

  class Parsers:
    def __init__(self):
      self.command_name = None
      self.argument = QuotedString(quoteChar="'",unquoteResults=True) | QuotedString(quoteChar='"',unquoteResults=True) | Word(printables)

    @property
    def command_string(self):
      parser = Suppress(Literal("#")) + self.command_name("name") + Optional(Suppress(Literal(":"))) + ZeroOrMore(self.argument)('arguments')
      return parser

  def __init__(self):
    self.have_pyparsing = have_pyparsing
    self.parsers = self.Parsers()
    self.aliases = {}

  def add_command(self,name,aliases=[]):
    if self.parsers.command_name is None:
       self.parsers.command_name = Literal(name)
    else:
       self.parsers.command_name = self.parsers.command_name ^ Literal(name)

    for alias in aliases:
       self.parsers.command_name = self.parsers.command_name ^ Literal(alias)
       self.aliases[alias] = name



  def parse(self,text):
    try:
      result = self.parsers.command_string.parseString(text,parseAll=True)
      ret = {}
      ret["name"] = self.aliases.get(result.name,result.name)
      ret["args"] = result.arguments
      return ret
    except:
      return None






import sys,tty,string
from collections import deque
import io
from .TerminalSession import *

class ScriptRecorder:

  def __init__(self,shell):
    self.shell = shell
    self.STDINFD = sys.stdin.fileno()
    self.exit_flag = False
    self.terminal = TerminalSession(self.shell)
    self.saved_terminal_settings = None
    self.lines = []

  def cleanup(self):
      try:
        self.terminal.stop()
      except: pass

      try:
        termios.tcsetattr(self.STDINFD, termios.TCSANOW, self.saved_terminal_settings)
      except: pass

  def get_bytes(self):
    return b'\n'.join(self.lines)

  def run(self):
    
    self.saved_terminal_settings = termios.tcgetattr(self.STDINFD)
    tty.setraw(self.STDINFD)

    def add_line(stack):
      out = io.BytesIO()
      while True:
        try:
          out.write(stack.popleft())
        except:
          break
      self.lines.append(out.getvalue())

    line = deque()
    while True:
      input = os.read(self.STDINFD,1024)
      if len(input) == 1 and input[0] in [3]:
        sys.exit(1)

      if len(input) == 1 and input[0] in [4]:
        add_line(line)
        break

      elif len(input) == 1 and input[0] in [127]: # backspace
        try:
            line.pop()
        except: pass

      elif len(input) == 1 and input[0] in [13]: # return
        add_line(line)

      else:
        line.append(input)

      self.terminal.send(input)

    termios.tcsetattr(self.STDINFD, termios.TCSANOW, self.saved_terminal_settings)





import os
from setuptools import setup, find_packages

__version__= '0.1.1'

setup(
  name = 'pygsc',
  packages = ['pygsc'],
  version = __version__,
  description = "Run command line demos from a script.",
  license='MIT',
  author = 'CD Clark III',
  author_email = 'clifton.clark@gmail.com',
  url = 'https://github.com/CD3/pygsc',
  download_url = f'https://github.com/CD3/pygsc/archive/{__version__}.tar.gz',
  install_requires = ['click','blessings','urwid','pygame'],
  entry_points='''
  [console_scripts]
  gsc=pygsc.cli:gsc
  gsc-record=pygsc.cli:gsc_record
  gsc-display-keycodes=pygsc.cli:display_keycodes
  gsc-monitor=pygsc.cli:gsc_monitor
  gsc-monitor-test-client=pygsc.cli:gsc_monitor_test_client
  gsc-monitor-test-server=pygsc.cli:gsc_monitor_test_server
  ''',
)
